@@ -28,10 +28,6 @@ struct sockopt {
 
 static int filan_streams_analyze(int fd, FILE *outfile);
 
-/* dirty workaround so we dont get an error on AIX when being linked with
-   libwrap */
-int allow_severity, deny_severity;
-
 /* global variables for configuring filan */
 bool filan_followsymlinks;
 bool filan_rawoutput;
@@ -39,6 +35,7 @@ bool filan_rawoutput;
 
 int sockoptan(int fd, const struct sockopt *optname, int socklay, FILE *outfile);
 int tcpan(int fd, FILE *outfile);
+int tcpan2(int fd, FILE *outfile);
 const char *getfiletypestring(int st_mode);
 
 static int printtime(FILE *outfile, time_t time);
@@ -82,17 +79,20 @@ int filan_file(const char *filename, FILE *outfile) {
    default:
       if ((fd =
 	   Open(filename,  O_RDONLY|O_NOCTTY|O_NONBLOCK
+#ifdef O_NOFOLLOW		
+		|(filan_followsymlinks?0:O_NOFOLLOW)
+#endif
 #ifdef O_LARGEFILE
 		|O_LARGEFILE
 #endif
 		, 0700))
 	  < 0) {
-	 Warn2("open(\"%s\", O_RDONLY|O_NOCTTY|O_NONBLOCK|O_LARGEFILE, 0700): %s",
+	 Warn2("open(\"%s\", O_RDONLY|O_NOCTTY|O_NONBLOCK|O_NOFOLLOW|O_LARGEFILE, 0700): %s",
 	       filename, strerror(errno));
       }
    }
      
-   result = filan_stat(&buf, fd, -1, outfile);
+   result = filan_stat(&buf, fd, -1, outfile, filename);
    fputc('\n', outfile);
    return result;
 }
@@ -122,7 +122,7 @@ int filan_fd(int fd, FILE *outfile) {
    }
    Debug2("fd %d is a %s", fd, getfiletypestring(buf.st_mode));
 
-   result = filan_stat(&buf, fd, fd, outfile);
+   result = filan_stat(&buf, fd, fd, outfile, NULL);
 
    if (result >= 0) {
       /* even more dynamic info */
@@ -215,7 +215,12 @@ int filan_stat(
 #else
 	       struct stat *buf
 #endif /* !HAVE_STAT64 */
-	       , int statfd, int dynfd, FILE *outfile) {
+	       , int statfd, int dynfd, FILE *outfile,
+	       const char *filename 	/* Linux does not (yet) provide an
+					   freadlink system call, so we need
+					   the original name for readlink in
+					   case it is a symlink */
+	       ) {
    char stdevstr[8];
 
    /* print header */
@@ -383,7 +388,11 @@ int filan_stat(
       break;
    case (S_IFREG):	/* 8, regular file */
       break;
+#ifdef S_IFLNK
    case (S_IFLNK):	/* 10, symbolic link */
+      /* we wait for freadlink() sytem call */
+      break;
+#endif /* S_IFLNK */
       break;
 #ifdef S_IFSOCK
    case (S_IFSOCK): /* 12, socket */
@@ -396,6 +405,19 @@ int filan_stat(
       break;
 #endif /* S_IFSOCK */
    }
+  } else {
+     switch (buf->st_mode&S_IFMT) {
+#ifdef S_IFLNK
+     case (S_IFLNK):	/* 10, symbolic link */
+	{
+	   char linktarget[PATH_MAX+1];
+	   memset(linktarget, 0, PATH_MAX+1);
+	   Readlink(filename, linktarget, PATH_MAX);
+	   fprintf(outfile, "LINKTARGET=%s", linktarget);
+	}
+	break;
+#endif /* S_IFLNK */
+     }
   }
    /* ioctl() */
    return 0;
@@ -531,20 +553,18 @@ int sockan(int fd, FILE *outfile) {
 #define FILAN_NAMELEN 256
    socklen_t optlen;
    int result /*0, i*/;
-   static const char *socktypes[] = {
-      "undef", "STREAM", "DGRAM", "RAW", "RDM",
-      "SEQPACKET", "undef", "undef", "undef", "undef", 
-      "PACKET", "undef" } ;
    char nambuff[FILAN_NAMELEN];
    /* in Linux these optcodes are 'enum', but on AIX they are bits! */
    static const struct sockopt sockopts[] = {
       {SO_DEBUG, "DEBUG"},
       {SO_REUSEADDR, "REUSEADDR"},
-      {SO_TYPE, "TYPE"},
-      {SO_ERROR, "ERROR"},
-#ifdef SO_PROTOTYPE
+#ifdef SO_PROTOCOL
+      {SO_PROTOCOL, "PROTOCOL"},
+#elif defined(SO_PROTOTYPE)
       {SO_PROTOTYPE, "PROTOTYPE"},
 #endif
+      {SO_TYPE, "TYPE"},
+      {SO_ERROR, "ERROR"},
       {SO_DONTROUTE, "DONTROUTE"},
       {SO_BROADCAST, "BROADCAST"},
       {SO_SNDBUF, "SNDBUF"},
@@ -619,8 +639,12 @@ int sockan(int fd, FILE *outfile) {
       Debug4("getsockopt(%d, SOL_SOCKET, SO_TYPE, %p, {"F_socklen"}): %s",
 	     fd, optval.c, optlen, strerror(errno));
    } else {
+#     define TYPENAMEMAX 16
+      char typename[TYPENAMEMAX];
+      sockettype(*optval.i, typename, sizeof(typename));
+      
       Debug3("fd %d: socket of type %d (\"%s\")", fd, *optval.i,
-	  socktypes[*optval.i]);
+	  typename);
    }
 
    optname = sockopts; while (optname->so) {
@@ -747,6 +771,9 @@ int ipan(int fd, FILE *outfile) {
 #ifdef IP_RECVTOS
       {IP_RECVTOS,      "IP_RECVTOS"},
 #endif
+#ifdef IP_TRANSPARENT
+      {IP_TRANSPARENT,  "IP_TRANSPARENT"},
+#endif
 #ifdef IP_MTU
       {IP_MTU,          "IP_MTU"},
 #endif
@@ -761,22 +788,29 @@ int ipan(int fd, FILE *outfile) {
 #endif
       {0, NULL} } ;
    const struct sockopt *optname;
-   int opttype;
-   socklen_t optlen = sizeof(opttype);
+   int optproto;
+   socklen_t optlen = sizeof(optproto);
    
    optname = ipopts; while (optname->so) {
       sockoptan(fd, optname, SOL_IP, outfile);
       ++optname;
    }
-   /* want to pass the fd to the next layer protocol. dont know how to get the
-      protocol number from the fd? use TYPE to identify TCP. */
-   if (Getsockopt(fd, SOL_SOCKET, SO_TYPE, &opttype, &optlen) >= 0) {
-      switch (opttype) {
+   /* want to pass the fd to the next layer protocol. */
+#if defined(SO_PROTOCOL) || defined(SO_PROTOTYPE)
+   if (Getsockopt(fd, SOL_SOCKET,
+#ifdef SO_PROTOCOL
+		  SO_PROTOCOL,
+#elif defined(SO_PROTOTYPE)
+		  SO_PROTOTYPE,
+#endif
+		  &optproto, &optlen) >= 0) {
+      switch (optproto) {
 #if WITH_TCP
-      case SOCK_STREAM: tcpan(fd, outfile); break;
+      case IPPROTO_TCP: tcpan(fd, outfile); break;
 #endif
       }
    }
+#endif /* defined(SO_PROTOCOL) || defined(SO_PROTOTYPE) */
    return 0;
 }
 #endif /* WITH_IP */
@@ -887,10 +921,46 @@ int tcpan(int fd, FILE *outfile) {
       sockoptan(fd, optname, SOL_TCP, outfile);
       ++optname;
    }
+
+#ifdef TCP_INFO
+   tcpan2(fd, outfile);
+#endif
    return 0;
 }
 #endif /* WITH_TCP */
 
+#if WITH_TCP && defined(TCP_INFO)
+
+int tcpan2(int fd, FILE *outfile) {
+   struct tcp_info tcpinfo;
+   socklen_t tcpinfolen = sizeof(tcpinfo);
+   int result;
+
+   result = Getsockopt(fd, SOL_TCP, TCP_INFO, &tcpinfo, &tcpinfolen);
+   if (result < 0) {
+      Debug4("getsockopt(%d, SOL_TCP, TCP_INFO, %p, {"F_Zu"}): %s",
+	     fd, &tcpinfo, sizeof(tcpinfo), strerror(errno));
+      return -1;
+   }
+   fprintf(outfile, "%s={%u}\t", "TCPI_STATE",		tcpinfo.tcpi_state);
+#if 0  /* on BSD these components are prefixed with __ - I get tired... */
+   fprintf(outfile, "%s={%u}\t", "TCPI_CA_STATE", 	tcpinfo.tcpi_ca_state);
+   fprintf(outfile, "%s={%u}\t", "TCPI_RETRANSMITS", 	tcpinfo.tcpi_retransmits);
+   fprintf(outfile, "%s={%u}\t", "TCPI_PROBES", 	tcpinfo.tcpi_probes);
+   fprintf(outfile, "%s={%u}\t", "TCPI_BACKOFF", 	tcpinfo.tcpi_backoff);
+#endif
+   fprintf(outfile, "%s={%u}\t", "TCPI_OPTIONS", 	tcpinfo.tcpi_options);
+   fprintf(outfile, "%s={%u}\t", "TCPI_SND_WSCALE", 	tcpinfo.tcpi_snd_wscale);
+   fprintf(outfile, "%s={%u}\t", "TCPI_RCV_WSCALE", 	tcpinfo.tcpi_rcv_wscale);
+   //fprintf(outfile, "%s={%u}\t", "TCPI_DELIVERY_RATE_APP_LIMITED", tcpinfo.tcpi_delivery_rate_app_limited);
+   //fprintf(outfile, "%s={%u}\t", "TCPI_FASTOPEN_CLIENT_FAIL", tcpinfo.tcpi_fastopen_client_fail);
+   // fprintf(outfile, "%s={%u}\t", "TCPI_", tcpinfo.tcpi_);
+
+   return 0;
+}
+
+#endif /* WITH_TCP */
+
 
 #if _WITH_SOCKET
 int sockoptan(int fd, const struct sockopt *optname, int socklay, FILE *outfile) {
