@@ -1,5 +1,5 @@
 /* source: xioread.c */
-/* Copyright Gerhard Rieger 2001-2010 */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* this is the source of the extended read function */
@@ -42,6 +42,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 
    if (pipe->readbytes) {
       if (pipe->actbytes == 0) {
+	 Info("xioread(): readbytes consumed, inserting EOF");
 	 return 0;	/* EOF by count */
       }
 
@@ -118,6 +119,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
       socklen_t fromlen = sizeof(from);
       char infobuff[256];
       char ctrlbuff[1024];	/* ancillary messages */
+      int rc;
 
       msgh.msg_name = &from;
       msgh.msg_namelen = fromlen;
@@ -127,9 +129,16 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 #if HAVE_STRUCT_MSGHDR_MSGCONTROLLEN
       msgh.msg_controllen = sizeof(ctrlbuff);
 #endif
-      if (xiogetpacketsrc(pipe->fd, &msgh) < 0) {
-	 return -1;
-      }
+
+      while ((rc = xiogetpacketsrc(pipe->fd, &msgh,
+			  MSG_PEEK
+#ifdef MSG_TRUNC
+			  |MSG_TRUNC
+#endif
+				   )) < 0 &&
+	     errno == EINTR) ;
+      if (rc < 0)  return -1;
+
       do {
 	 bytes =
 	    Recvfrom(pipe->fd, buff, bufsiz, 0, &from.soa, &fromlen);
@@ -221,9 +230,12 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
       }
 
       switch(from.soa.sa_family) {
+#if HAVE_STRUCT_IP
 	 int headlen;
+#endif /* HAVE_STRUCT_IP */
 #if WITH_IP4
       case AF_INET:
+#if HAVE_STRUCT_IP
 	 if (pipe->dtype & XIOREAD_RECV_SKIPIP) {
 	    /* IP4 raw sockets include the header when passing a packet to the
 	       application - we don't need it here. */
@@ -240,6 +252,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 	       bytes -= headlen;
 	    }
 	 }
+#endif /* HAVE_STRUCT_IP */
 	 break;
 #endif
 #if WITH_IP6
@@ -309,6 +322,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
       char infobuff[256];
       struct msghdr msgh = {0};
       char ctrlbuff[1024];	/* ancillary messages */
+      int rc;
 
       socket_init(pipe->para.socket.la.soa.sa_family, &from);
       /* get source address */
@@ -320,9 +334,15 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 #if HAVE_STRUCT_MSGHDR_MSGCONTROLLEN
       msgh.msg_controllen = sizeof(ctrlbuff);
 #endif
-      if (xiogetpacketsrc(pipe->fd, &msgh) < 0) {
-	 return -1;
-      }
+      while ((rc = xiogetpacketsrc(pipe->fd, &msgh,
+			  MSG_PEEK
+#ifdef MSG_TRUNC
+			  |MSG_TRUNC
+#endif
+				   )) < 0 &&
+	     errno == EINTR) ;
+      if (rc < 0)  return -1;
+
       xiodopacketinfo(&msgh, true, false);
       if (xiocheckpeer(pipe, &from, &pipe->para.socket.la) < 0) {
 	 Recvfrom(pipe->fd, buff, bufsiz, 0, &from.soa, &fromlen);  /* drop */
@@ -357,9 +377,12 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
       }
 
       switch(from.soa.sa_family) {
+#if HAVE_STRUCT_IP
 	 int headlen;
+#endif /* HAVE_STRUCT_IP */
 #if WITH_IP4
       case AF_INET:
+#if HAVE_STRUCT_IP
 	 if (pipe->dtype & XIOREAD_RECV_SKIPIP) {
 	    /* IP4 raw sockets include the header when passing a packet to the
 	       application - we don't need it here. */
@@ -376,6 +399,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 	       bytes -= headlen;
 	    }
 	 }
+#endif /* HAVE_STRUCT_IP */
 	 break;
 #endif
 #if WITH_IP6
